/* ============ 1/f.c ============== */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* ==== MIDI Interfacing stuff ===== */

#include <MIDI.h>

#define clientID '1/f '
#define timePortID '1/fa'
#define inputPortID '1/fi'
#define outputPortID '1/fo'

static short int tpr, ipr, opr;
static MIDIPortParams iO, iI, iT;
static MIDIPortInfoHdl pi;
static MIDIPacket mP;
static Boolean tR = false;
Boolean emt = false;
static long int fet = 0;

static OSErr Signal(OSErr err)
{
  if (err != noErr)
    {
      fprintf(stderr, "Error %d occurred\n", err);
    }
}

static pascal short mrh(MIDIPacketPtr mp, long int prevA5)
{
  register long int currA5;
  register unsigned char flag;

  asm
    {
      move.l a5, currA5
      move.l prevA5, a5
    }
  flag = mp->flags & midiContMask;
  if (emt)
    {
      (void) MIDIWritePacket(opr, mp);
    }
  if (flag == midiNoCont || flag == midiEndCont)
    {
      asm
	{
	  move.l currA5, a5
	}
      return (midiNoMorePacket);
    }
  asm
    {
      move.l currA5, a5
    }
  return (midiMorePacket);
}

static void InitializeMIDIManager(Handle icon)
{
  int i, j;
  MIDIIDListHdl clients, clientPorts;
  int clientNr;
  long int MIDIMgrVerNum = 0;

    /* Check for MIDI Manager: if you don't have SndDispVersion, 
     * delete the next 5 rules */ 

  if ((MIDIMgrVerNum = SndDispVersion(midiToolNum)) == 0)
    {
      Signal(-262);
      ExitToShell();
    }
  Signal(MIDISignIn(clientID, 0, icon, "\pArpegg"));
  iT.portID = timePortID; iT.portType = midiPortTypeTime;
  iT.timeBase = 0; iT.offsetTime = 0; iT.readHook = NULL;
  asm { move.l a5, iT.refCon }
  iT.initClock.sync = midiInternalSync; iT.initClock.curTime = 0;
  iT.initClock.format = midiFormatMSec; strcpy((char *) iT.name, "Time"); 
  CtoPstr(iT.name); 
  Signal(MIDIAddPort(clientID, 1024, &tpr, &iT));
  iI.portID = inputPortID; iI.portType = midiPortTypeInput;
  iI.timeBase = tpr; iI.offsetTime = 0;
  iI.readHook = (Ptr) mrh; asm { move.l a5, iI.refCon }
  iI.initClock.sync = midiInternalSync; iI.initClock.curTime = 0;
  iI.initClock.format = midiFormatMSec; strcpy((char *) iI.name, "Input");
  CtoPstr(iI.name); Signal(MIDIAddPort(clientID, 1024, &ipr, &iI));
  iO.portID = outputPortID; iO.portType = midiPortTypeOutput;
  iO.timeBase = tpr; iO.offsetTime = 0;
  iO.readHook = NULL; iO.refCon = outputPortID;
  iO.initClock.sync = midiInternalSync; iO.initClock.curTime = 0;
  iO.initClock.format = midiFormatMSec; strcpy((char *) iO.name, "Output"); CtoPstr(iO.name);
  Signal(MIDIAddPort(clientID, 1024, &opr, &iO));
  MIDIStartTime(tpr);
  clients = MIDIGetClients();
  for (clientNr = -1, i = 0; i != (*clients)->numIDs; i++)
    {
      if ((*clients)->list[i] == 'amdr')
	{
	  clientNr = i;
	  clientPorts = MIDIGetPorts((*clients)->list[clientNr]);
	  for (j = 0; j != (*clientPorts)->numIDs; j++)
	    {
	      pi = MIDIGetPortInfo((*clients)->list[clientNr], (*clientPorts)->list[j]);
	      if ((*pi)->portType == midiPortTypeTime)
		{
		  Signal(MIDIConnectTime((*clients)->list[clientNr], (*clientPorts)->list[j], clientID, timePortID));
		}
	      else if ((*pi)->portType == midiPortTypeInput)
		{
		  Signal(MIDIConnectData(clientID, outputPortID, (*clients)->list[clientNr], (*clientPorts)->list[j]));
		}
	      else if ((*pi)->portType == midiPortTypeOutput)
		{
		  Signal(MIDIConnectData((*clients)->list[clientNr], (*clientPorts)->list[j], clientID, inputPortID));
		}
	    }
	  break;
	}
    }
  if (clientNr == -1)
    {
      Signal(10011);
    }
}

static void AllNotesOff(void)
{
    int i;

    if (tR)
    {
	tR = false;
	MIDIStopTime(tpr);
    }
    mP.len = 9;
    mP.flags = midiTimeStampCurrent | midiMsgType | midiNoCont;
    mP.tStamp = 0;
    for (i = 0; i != 16; i++)
    {
	mP.data[0] = 0xB0 + i;
	mP.data[1] = 0x7E;
	mP.data[2] = 0x00;
	(void) MIDIWritePacket(opr, &mP);
	mP.data[1] = 0x7B;
	(void) MIDIWritePacket(opr, &mP);
	mP.data[1] = 0x79;
	(void) MIDIWritePacket(opr, &mP);
    }
}

static void FinalizeMIDIManager(void)
{
    AllNotesOff();
    MIDISignOut(clientID);
}

static void TransmitMidi(short int portRefnum, Byte *message, unsigned int length, long int time)
{
    if (time > fet)
    {
	fet = time;
    }
    memcpy(mP.data, message, length);
    mP.len = length + 6;
    mP.flags = midiTimeStampValid | midiMsgType | midiNoCont;
    mP.tStamp = time;
    Signal(MIDIWritePacket(portRefnum, &mP));
}

/* Plays the given note at the given channel with the given velocity at time cTime,
   sends a note off length ticks (= 1 ms) later */
static void PlayNote(long int cTime, Byte channel, Byte note, Byte velocity, int length)
{
    Byte message[3];

    if (note == 0 || note > 127 || channel > 15 || velocity == 0 || velocity > 127) return;
    message[0] = 0x90 | (Byte) channel;
    message[1] = note;
    message[2] = velocity;
    TransmitMidi(opr, message, 3, cTime);
    message[0] = 0x80 | (Byte) channel;
    TransmitMidi(opr, message, 3, cTime + length);
}

/* ===== Music stuff ===== */

/* C major, full octave, in MIDI note numbers */

static short int scale[] = {60, 62, 64, 65, 67, 69, 71, 72};

/* You can change this scale to anything you like, e.g.
   {61, 63, 66, 68, 70, 73} will give you a pentatonic
   scale, and {60, 62, 63, 65, 67, 68, 70, 72} C minor.
   Make sure the scale is complete (e.g., the C major
   scale runs from C to C, including both). The twelve tone
   scale is {60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72}
 */

/* Constrains notes to a scale, given above */
static void PlayScaleNote(int note, int velocity, int length)
{
    Byte nNote;
    int scaleLength = sizeof(scale) / sizeof(scale[0]) - 1;
    int aNote = note % scaleLength;

    if (aNote < 0)
    {
	aNote += scaleLength;
    }
    nNote = scale[aNote] + (note - aNote) / scaleLength * (scale[scaleLength] - scale[0]);
    PlayNote(fet, 0, nNote, velocity, length);
}

/* Generates random number between low and high */
static int rnd(int low, int high)
{
    return (((long int) rand() * (high - low + 1)) / (RAND_MAX + 1) + low);
}

static int nrGen;
static int range;
static int lastLength = 8;
static int mode;

/* This is the 1/f generator */
static void GenerateNote(int nr)
{
    int prev = nr - 1;
    int note = 0;
    int i = 0, b = 1;
    int length;

    while (i != nrGen)
    {
	if ((nr & b) != (prev & b))
	{
	    note += rnd(-range, range);
	}
	i++;
	b <<= 1;
    }
    length = rand() & 0x7;
    b = 1;
    while (b != 8 && (length & b) == 0)
    {
	b <<= 1;
    }
    PlayScaleNote(note, 48 + lastLength * 2, b * 125);
    lastLength = b;
}

/* You can specify:
   range: each generator will generate a value between (and including) -range and +range
	  keep it low, otherwise the melody will become jumpy
   nrgen: the number of generators (8 or 9 are nice numbers)
   length: the number of notes to be generated (take 2^nrgen)
   mode: 0 = normal 1/f, the generator steps linearly through the
                         values 1..length
         1 = random 1/f, the generator steps in random order through
                         the values 1..2^nrGen
*/
int main(void)
{
    int i = 0;
    int len;

    _atexit(FinalizeMIDIManager);
    InitializeMIDIManager(icon);
    srand(TickCount());
    while ((printf("Range = "), scanf("%d", &range)) == 1 &&
	   (printf("NrGen = "), scanf("%d", &nrGen)) == 1 &&
	   (printf("Length = "), scanf("%d", &len)) == 1 &&
	   (printf("Mode = "), scanf("%d", &mode)) == 1)
    {       
	fet = MIDIGetCurTime(opr) + 50;
	i = 0; lastLength = 8;
	while (i != len)
	{
	    if (MIDIGetCurTime(opr) + 600 > fet)
	    {
		i++;
		GenerateNote(mode == 0? i: rnd(1, 1 << nrGen));
	    }
	}
	while (MIDIGetCurTime(opr) <= fet)
	{
	    /* Just wait to prevent overflow... */
	}
    }
}

--
	Theo Vosse
	----------
	Unit for Experimental Psychology
	University of Leiden
	The Netherlands

